@@ -1,4 +1,4 @@
-/* Copyright (C) 2007-2016  B.A.T.M.A.N. contributors:
+/* Copyright (C) 2007-2017  B.A.T.M.A.N. contributors:
  *
  * Marek Lindner, Simon Wunderlich
  *
@@ -24,17 +24,14 @@
 #include <linux/version.h>
 #include_next <linux/compiler.h>
 
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 33)
+#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 19, 0)
 
-#define __always_unused			__attribute__((unused))
-#define __percpu
+#define READ_ONCE(x) ACCESS_ONCE(x)
 
-#endif /* < KERNEL_VERSION(2, 6, 33) */
+#define WRITE_ONCE(x, val) ({ \
+	ACCESS_ONCE(x) = (val); \
+})
 
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 36)
-
-#define __rcu
-
-#endif /* < KERNEL_VERSION(2, 6, 36) */
+#endif /* < KERNEL_VERSION(3, 19, 0) */
 
 #endif	/* _NET_BATMAN_ADV_COMPAT_LINUX_COMPILER_H_ */
